/* -*-Mode:C; tab-width:4; indent-tabs-mode:t; c-file-style:"stroustrup";-*- */
/* Main code for Project */
//#define ERROR_CHECK_LEVEL 2
#include <PalmOS.h>
#include <DateGlue.h>
#include <TxtGlue.h>
#include <FntGlue.h>
#include <TsmGlue.h>
#include <WinGlue.h>
#include "handera/Vga.h"
//#include "SonyCLIE.h"

#include "link.h"
#include "task.h"
#include "xb.h"
#include "progectdb.h"
#include "progectRsc.h"
#include "progect.h"
#include "flat.h"
#include "btm.h"
#include "beam.h"

// icon support
#include "icon.h"
// ToDo export support
#include "ToDoDB.h"
// Memo linking support
#include "MemoDB.h"
// address linking support
#include "AddressDB.h"
// Doc export support
#include "doc.h"
// Category support
#include "category.h"
// history support
#include "history.h"
// hotsync support
#include "aftersync.h"

#include "projlistMODEL.h"

// prototypes

// events

// app info block for Project List Database
typedef struct {
	AppInfoType appInfo;
	UInt8 appInfoVersion;
	UInt8 sortOrderType;
	UInt16 currentCategory;
	UInt16 lastSentinelValue;
} ProjectListAppInfoType;

// statics
static struct ProjectListMODELGlobals *PLM_Globals=NULL;

ProjectListEntry *ProjectListMODELGetRecord(UInt16 index)
{
MemHandle dbRecord;
ProjectListEntry *dbRecordStruct;

	dbRecord=DmQueryRecord(PLM_Globals->databaseRef,index);
	if (NULL == dbRecord)
	{
		// +++ FIX THIS +++ Output error in debug mode
		return NULL;
	}

	dbRecordStruct=MemHandleLock(dbRecord);
	if (NULL == dbRecordStruct)
	{
		// +++ FIX THIS +++ Output error in debug mode
		return NULL;
	}

	// make sure you MemPtrUnlock this...
	return dbRecordStruct;
}

void ProjectListMODELInit(void)
{
	ProjectListMODELInitINTERNAL(&PLM_Globals);
}

void ProjectListMODELInitINTERNAL(struct ProjectListMODELGlobals **plGlobals)
{
	const char *listDBName="lbPG-PROJECT_META_DATA";
	Err err;
	LocalID dbID;
	Boolean initializeMetaDataDatabase=false;
	MemHandle h;
	LocalID appInfoID;
	ProjectListAppInfoType *appInfoP;

	char deleteMe[40];

//	DEBUG1("About to init");

	// If we have already allocated arrays then just increase the refcount
	if (NULL != *plGlobals)
	{
	StrIToA(deleteMe, (*plGlobals)->refCount);
//	DEBUG2("Current ref count1 init",deleteMe);
		(*plGlobals)->refCount++;
	StrIToA(deleteMe, (*plGlobals)->refCount);
//	DEBUG2("Current ref count2 init",deleteMe);
		return;
	}

//	DEBUG1("About to create structs in init");

	*plGlobals=MemPtrNew(sizeof(struct ProjectListMODELGlobals));
	MemSet(*plGlobals,sizeof(struct ProjectListMODELGlobals),0);
	
	// have to start at one, because the deallocation routine assumes this
	(*plGlobals)->refCount=1;

	StrIToA(deleteMe, (*plGlobals)->refCount);
//	DEBUG2("Current ref count3 init",deleteMe);

	//
	// Open the database that lists all the projects we know about
	//
	dbID=DmFindDatabase((*plGlobals)->cardNumber,listDBName);
	if (NULL == dbID)
	{
		err=DmGetLastErr();
		if (dmErrCantFind == err)
		{
			// doesn't exist, okay, then try creating it
			err=DmCreateDatabase((*plGlobals)->cardNumber,listDBName,'lbPG','META',false);
			if (err)
			{
				// +++ FIX THIS +++
				// Failed to create
				// Bring up fatal error dialog, and quit program
				// +++ FIX THIS +++
				DEBUG1("Failed to create project Metadata database");
				return;
			}

		// let's try that find _again_
		dbID=DmFindDatabase((*plGlobals)->cardNumber,listDBName);
		// remember we need to create the AppInfoBlock, since the database was created
		// and will not have one.
		initializeMetaDataDatabase=true;
		}
	}
	if (NULL == dbID)
	{
		// +++ FIX THIS +++
		// Failed to find the database we need, even after we must have successfully
		// created it.
		// Bring up fatal error dialog, and quit program
		// +++ FIX THIS +++
		DEBUG1("Failed to open project Metadata database");
		return;
	}
	(*plGlobals)->databaseRef=DmOpenDatabase((*plGlobals)->cardNumber,dbID,dmModeReadWrite);
	if (NULL == (*plGlobals)->databaseRef)
	{
		// +++ FIX THIS +++
		// Failed to open the database we need
		// Bring up fatal error dialog, and quit program
		// +++ FIX THIS +++
		return;
	}

	//
	// Create the AppInfoBlock if required
	//
	if (initializeMetaDataDatabase)
	{
		UInt16 attributes;

		// set the backup bit
		DmDatabaseInfo((*plGlobals)->cardNumber,dbID,NULL,&attributes,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
		attributes |= dmHdrAttrBackup;
		
		h=DmNewHandle((*plGlobals)->databaseRef,sizeof(ProjectListAppInfoType));
		if (!h)
		{
			// +++ FIX THIS +++
			// Failed to create the AppInfo Block
			// Bring up fatal error dialog, and quit program
			// +++ FIX THIS +++
			return;
		}
		appInfoID=MemHandleToLocalID(h);
		// set the application info block
		err=DmSetDatabaseInfo((*plGlobals)->cardNumber,dbID,NULL,&attributes,NULL, NULL,NULL,NULL,NULL,&appInfoID,NULL,NULL,NULL);
		if (errNone == err)
		{
			appInfoP=(ProjectListAppInfoType *)MemHandleLock(h);
			// clear everything
			DmSet(appInfoP,0,sizeof(ProjectListAppInfoType),0);
			// set category info
			CategoryInitialize(&appInfoP->appInfo,ProjectDefaultCategoryList);
			// unlock
			MemPtrUnlock(appInfoP);
		}
		else
		{
			MemHandleFree(h);
			// +++ FIX THIS +++
			// Failed to create the AppInfo Block
			// Bring up fatal error dialog, and quit program
			// +++ FIX THIS +++
			return;
		}
	}
} // void ProjectListMODELInit(void)

void ProjectListMODELCleanUp(void)
{
	ProjectListMODELCleanupINTERNAL(&PLM_Globals);
}

void ProjectListMODELCleanupINTERNAL(struct ProjectListMODELGlobals **plGlobals)
{
	char deleteMe[40];

//	DEBUG1("About to de-init");

	StrIToA(deleteMe, (*plGlobals)->refCount);
//	DEBUG2("Current ref count1 Cleanup",deleteMe);

	if (--(*plGlobals)->refCount)
		return;

//	DEBUG1("Really going to delete");

	StrIToA(deleteMe, (*plGlobals)->refCount);
//	DEBUG2("Current ref count2 Cleanup",deleteMe);

	//
	// Close the project list database
	// 
	if ((*plGlobals)->databaseRef)
		DmCloseDatabase((*plGlobals)->databaseRef);
	(*plGlobals)->databaseRef=NULL;

	MemPtrFree(*plGlobals);
	*plGlobals=NULL;

//	DEBUG1("All done deleting");
	
} // void ProjectListMODELCleanUp(void)

UInt16 ProjectListMODELGetCurrentCategory(void)
{
	ProjectListAppInfoType *appInfoP;
	UInt16 currentCategory;

		// grab the application info block and get the category
		appInfoP=(ProjectListAppInfoType *)MemLocalIDToLockedPtr(DmGetAppInfoID(PLM_Globals->databaseRef),PLM_Globals->cardNumber);
		currentCategory=appInfoP->currentCategory;
		MemPtrUnlock(appInfoP);

		return currentCategory;
}

void ProjectListMODELSetCurrentCategory(UInt16 currentCategory)
{
	ProjectListAppInfoType *appInfoP;

		// grab the application info block and set the category
		appInfoP=(ProjectListAppInfoType *)MemLocalIDToLockedPtr(DmGetAppInfoID(PLM_Globals->databaseRef),PLM_Globals->cardNumber);
		DmWrite(appInfoP, OffsetOf(ProjectListAppInfoType, currentCategory), &currentCategory, sizeof(UInt16));
		MemPtrUnlock(appInfoP);
}

UInt16 ProjectListMODELGetRecordCategory(UInt16 record)
{
	UInt16 attr;

	DmRecordInfo (PLM_Globals->databaseRef, record, &attr, NULL, NULL);
	attr &= dmRecAttrCategoryMask;
	return attr;
}

void ProjectListMODELSetRecordCategory(UInt16 record,UInt16 newCategory)
{
	UInt16 attr;

	// If the category is All then set the category to unfiled.
	DmRecordInfo (PLM_Globals->databaseRef, record, &attr, NULL, NULL);
	attr &= ~dmRecAttrCategoryMask;
	attr |= ((dmAllCategories == newCategory) ? dmUnfiledCategory :
			 newCategory) | dmRecAttrDirty;
	DmSetRecordInfo (PLM_Globals->databaseRef, record, &attr, NULL);
}

static UInt16 GetLastSentinelValueINTERNAL(struct ProjectListMODELGlobals **plGlobals)
{
	ProjectListAppInfoType *appInfoP;
	UInt16 lastSentinelValue;

		// grab the application info block and get the category
		appInfoP=(ProjectListAppInfoType *)MemLocalIDToLockedPtr(DmGetAppInfoID((*plGlobals)->databaseRef),(*plGlobals)->cardNumber);
		lastSentinelValue=appInfoP->lastSentinelValue;
		MemPtrUnlock(appInfoP);

		return lastSentinelValue;
}

static void SetLastSentinelValueINTERNAL(struct ProjectListMODELGlobals **plGlobals, UInt16 lastSentinelValue)
{
	ProjectListAppInfoType *appInfoP;

		// grab the application info block and set the category
		appInfoP=(ProjectListAppInfoType *)MemLocalIDToLockedPtr(DmGetAppInfoID((*plGlobals)->databaseRef),(*plGlobals)->cardNumber);
		DmWrite(appInfoP, OffsetOf(ProjectListAppInfoType, lastSentinelValue), &lastSentinelValue, sizeof(UInt16));
		MemPtrUnlock(appInfoP);
}


DmOpenRef ProjectListMODELGetDBRef(void)
{
	return PLM_Globals->databaseRef;

} // DmOpenRef ProjectListMODELGetDBRef 

void updateAddProjectModelEntryINTERNAL(struct ProjectListMODELGlobals **plGlobals, ProjectListEntry *newEntry)
{
UInt16 recordNumber;
MemHandle dbRecord;
ProjectListEntry *dbRecordStruct;
UInt16 sentinelValue=GetLastSentinelValueINTERNAL(plGlobals);

	if (FindInSortedDatabase((*plGlobals)->databaseRef,CompareProjectEntryByCardProjectName,newEntry,&recordNumber))
	{
		//
		// Already exists
		//
		dbRecord=DmGetRecord((*plGlobals)->databaseRef,recordNumber);
		if (NULL == dbRecord)
		{
			// +++ FIX THIS +++ Output error in debug mode
			return;
		}
		else
		{
			dbRecordStruct=MemHandleLock(dbRecord);
			if (NULL == dbRecordStruct)
			{
				// +++ FIX THIS +++ Output error in debug mode
				return;
			}
			//
			// update the sentinel value so we know that we've seen it during this scan
			//
			DmWrite(dbRecordStruct,OffsetOf(ProjectListEntry,lastSentinelValue),&sentinelValue,sizeof(sentinelValue));

			MemHandleUnlock(dbRecord);

			DmReleaseRecord((*plGlobals)->databaseRef,recordNumber,true); // true means record was modified
		}
	}
	else
	{
		//
		// Have to create
		//
		//
		dbRecord=DmNewRecord((*plGlobals)->databaseRef,&recordNumber,sizeof(ProjectListEntry));
		if (NULL == dbRecord)
		{
			DEBUG1("Error creating new record in Project List Database");
			return;
		}
		else
		{
			dbRecordStruct=MemHandleLock(dbRecord);
			if (NULL == dbRecordStruct)
			{
				// +++ FIX THIS +++ Output error in debug mode
				return;
			}

			//
			// update the sentinel value so we know that we've seen it during this scan
			//
			newEntry->lastSentinelValue=sentinelValue;

			//
			// write out the new record
			// 
			DmWrite(dbRecordStruct,0,newEntry,sizeof(ProjectListEntry));

			MemHandleUnlock(dbRecord);

			DmReleaseRecord((*plGlobals)->databaseRef,recordNumber,true); // true means record was modified
		}
	}
}

void ProjectListMODELUpdate(void)
{
	ProjectListMODELUpdateINTERNAL(&PLM_Globals);
}

void ProjectListMODELUpdateINTERNAL(struct ProjectListMODELGlobals **plGlobals)
{
	Boolean newSearch=true;
	DmSearchStateType stateInfo;
	ProjectListEntry newEntry;
	LocalID dbID;
	Err err;

	UInt16 i;
	MemHandle dbRecord;
	ProjectListEntry *dbRecordStruct;
	UInt16 sentinelValue;
	Boolean deleteTheRecord;

	//
	// increment the sentinel value so we know when an item has not been seen and can be deleted
	// I guess you could call our method mark and sweep, we go through all the databases, adding them
	// to our database if we have not seen them (with the latest sentinel value), otherwise we simply 
	// update the sentinel value. At the end of this function we delete any items that have a sentinel 
	// value different than the last one.
	//
	sentinelValue=GetLastSentinelValueINTERNAL(plGlobals)+1;
	SetLastSentinelValueINTERNAL(plGlobals,sentinelValue);

	// zero out the fields of the ProjectListEntry for starters
	MemSet(&newEntry,sizeof(ProjectListEntry),0);

	//
	// Go through all the databases on the device and put them in the database
	//
	do 
	{
		// start the search
		err=DmGetNextDatabaseByTypeCreator(newSearch,&stateInfo,'DATA',CREATOR,false,&newEntry.cardNumber,&dbID);
		// flip state for API
		newSearch=false;

		// if we found something
		if (errNone == err)
		{
			// grab database name
			DmDatabaseInfo(newEntry.cardNumber, dbID, newEntry.databaseName, NULL, NULL, NULL, NULL, NULL,NULL, NULL, NULL, NULL, NULL);

			// if the name is prepended (new databases since V0.10), then remove the extra
			if (StrStr(newEntry.databaseName, gPrepend) == newEntry.databaseName)
				MemMove(newEntry.databaseName, newEntry.databaseName + StrLen(gPrepend),StrLen(newEntry.databaseName+StrLen(gPrepend)) + 1);

			// record the project in our database
			updateAddProjectModelEntryINTERNAL(plGlobals,&newEntry);
		}
	}
	while (errNone == err);

	//
	// Now we go about deleting databases that have dissapeared for some reason, this would typically be during a HotSync
	// or if some other program has deleted the database. Since we should delete entries in here ourselves when we delete them,
	// ditto for database creation and renaming.
	//

	// Go backwards so that deleting an entry does not change the numbering of the records we have not seen yet
	for (i=DmNumRecords((*plGlobals)->databaseRef);i>0;i--)
	{
		deleteTheRecord=false;	

		dbRecord=DmQueryRecord((*plGlobals)->databaseRef,i-1);
		if (NULL == dbRecord)
		{
			// +++ FIX THIS +++ Output error in debug mode
			return;
		}
		else
		{
			dbRecordStruct=MemHandleLock(dbRecord);
			if (NULL == dbRecordStruct)
			{
				// +++ FIX THIS +++ Output error in debug mode
				return;
			}
			//
			// If the sentinel's don't match, this project has gone away
			//
			if (dbRecordStruct->lastSentinelValue != sentinelValue)
				deleteTheRecord=true;
			
			MemHandleUnlock(dbRecord);

			if (true == deleteTheRecord)
				DmRemoveRecord((*plGlobals)->databaseRef,i-1);
		}
	}
}

// returns true if record is in database (recordNumber contains location)
// returns false if record not there (recordNumber contains insertion pos.).
Boolean FindInSortedDatabase(DmOpenRef db, DmComparF *comparFunc, void *findRec, UInt16 *recordNumber)
{
	Boolean     foundIt = false;
	   
	*recordNumber = DmFindSortPosition(db, findRec, NULL, comparFunc, 0);
	   
	if (*recordNumber > 0) 
	{
	MemHandle   h;
	void	*potentialMatchRec;

		// matching record would be at *recordNumber - 1
		h = DmQueryRecord(db, *recordNumber - 1);
		potentialMatchRec = MemHandleLock(h);
		if ((*comparFunc)(findRec, potentialMatchRec, 0, NULL, NULL, NULL) == 0)
		{
			foundIt = true;
			(*recordNumber)--;
		}

	MemHandleUnlock(h);
	}

	return foundIt;
}


// Function to compare two ProjectEntry structures by card number, then by project name
Int16 CompareProjectEntryByCardProjectName(void *r1, void *r2,Int16 unused1, SortRecordInfoPtr unused2, SortRecordInfoPtr unused3, MemHandle unused4)
{
	ProjectListEntry *rec1=(ProjectListEntry *)r1;
	ProjectListEntry *rec2=(ProjectListEntry *)r2;
	Int16 result;

	result = rec1->cardNumber-rec2->cardNumber;
	if (0 == result)
		result = StrCompare(rec1->databaseName,rec2->databaseName);
	return result;
}

/***********************************************************************
 *
 * FUNCTION:    ToolsSeekRecord
 *
 * DESCRIPTION: Given the index of a to do record, this routine scans
 *              forewards or backwards for displayable to do records.
 *
 * PARAMETERS:  indexP  - pointer to the index of a record to start from;
 *                        the index of the record sought is returned in
 *                        this parameter.
 *
 *              offset  - number of records to skip:
 *                           0 - mean seek from the current record to the
 *                             next display record, if the current record is
 *                             a display record, its index is retuned.
 *                         1 - mean seek foreward, skipping one displayable
 *                             record
 *                        -1 - means seek backwards, skipping one
 *                             displayable record
 *
 *
 * RETURNED:    false is return if a displayable record was not found.
 *
 ***********************************************************************/
Boolean ProjectListMODELSeekRecord (UInt16 * indexP, Int16 offset, Int16 direction)
{
	DmSeekRecordInCategory (PLM_Globals->databaseRef, indexP, offset, direction, ProjectListMODELGetCurrentCategory());
	if (DmGetLastErr()) return (false);

	return (true);
}

void ProjectListModelAddProject(UInt16 cardNo, const char *dbName)
{
UInt16 recordNumber;
ProjectListEntry newEntry;

	MemSet(&newEntry,sizeof(ProjectListEntry),0);

	newEntry.cardNumber=cardNo;
	SafeCopy(newEntry.databaseName,dbName,sizeof(newEntry.databaseName));
	updateAddProjectModelEntryINTERNAL(&PLM_Globals,&newEntry);

	if (FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,&newEntry,&recordNumber))
		ProjectListMODELSetRecordCategory(recordNumber,ProjectListMODELGetCurrentCategory());
}

void ProjectListModelRenameProject(UInt16 cardNo, const char *dbNameOld,const char *dbNameNew)
{
UInt16 recordNumber;
UInt16 category;
ProjectListEntry findEntry;
ProjectListEntry *currentEntry;
MemHandle currentEntryHandle;


	MemSet(&findEntry,sizeof(ProjectListEntry),0);

	findEntry.cardNumber=cardNo;
	SafeCopy(findEntry.databaseName,dbNameOld,sizeof(findEntry.databaseName));

	if (FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,&findEntry,&recordNumber))
	{
		category=ProjectListMODELGetRecordCategory(recordNumber);
		DmDetachRecord(PLM_Globals->databaseRef,recordNumber,&currentEntryHandle);
		currentEntry=MemHandleLock(currentEntryHandle);
		if (NULL == currentEntry)
		{
			DEBUG1("Unable to rename project");
			return;
		}

		DmWrite(currentEntry, OffsetOf(ProjectListEntry, databaseName), dbNameNew, sizeof(currentEntry->databaseName));
		
		if(!FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,currentEntry,&recordNumber))
		{
			MemPtrUnlock(currentEntry);
			DmAttachRecord(PLM_Globals->databaseRef,&recordNumber,currentEntryHandle,NULL);
			ProjectListMODELSetRecordCategory(recordNumber,category);
		}
		else
		{
			MemPtrUnlock(currentEntry);
			DEBUG1("Unable to rename project 2");
		}
	}
}

void ProjectListModelDeleteProject(UInt16 cardNo, const char *dbName)
{
UInt16 recordNumber;
ProjectListEntry newEntry;


	MemSet(&newEntry,sizeof(ProjectListEntry),0);

	newEntry.cardNumber=cardNo;
	SafeCopy(newEntry.databaseName,dbName,sizeof(newEntry.databaseName));

	if (FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,&newEntry,&recordNumber))
		DmRemoveRecord(PLM_Globals->databaseRef,recordNumber);
}

UInt16 ProjectListModelGetProjectCategory(UInt16 cardNo, const char *dbName)
{
UInt16 recordNumber;
ProjectListEntry newEntry;


	MemSet(&newEntry,sizeof(ProjectListEntry),0);

	newEntry.cardNumber=cardNo;
	SafeCopy(newEntry.databaseName,dbName,sizeof(newEntry.databaseName));
	if (FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,&newEntry,&recordNumber))
		return ProjectListMODELGetRecordCategory(recordNumber);

	return dmUnfiledCategory;
}


void ProjectListModelSetProjectCategory(UInt16 cardNo, const char *dbName, UInt16 newCategory)
{
UInt16 recordNumber;
ProjectListEntry newEntry;


	MemSet(&newEntry,sizeof(ProjectListEntry),0);

	newEntry.cardNumber=cardNo;
	SafeCopy(newEntry.databaseName,dbName,sizeof(newEntry.databaseName));

	if (FindInSortedDatabase(PLM_Globals->databaseRef,CompareProjectEntryByCardProjectName,&newEntry,&recordNumber))
		ProjectListMODELSetRecordCategory(recordNumber,newCategory);
}

